# parse the .tex file, stop on various errors, and parse the top-level structure
# Arguments:
#  tex: the tex file contents (using shinyTex rules)
#  fname: the main .tex filename (for reporting errors)
# Value:
#  a list containing
#    cleanTex: the cleaned up tex file
#    defs: a named list with all "\def" data with some defaults added in as needed
#    mainStartEnd: a matrix of the top-level structure with one row per main element,
#      and columns for the starting and ending line numbers
#    tabTitles: the titles of all Shiny tabs 
#    tabbing: a logical specifying if any tabs are in the page
#
parseTex = function(tex, fname) {
  ### Initial cleanup ###
  # convert R comments from %# to #
  tex = gsub("%#", "#", tex, fixed=TRUE)
  # blank out whole-line tex comments
  temp = grep("^[[:blank:]]*%", tex)
  tex[temp] = ""
  # remove other tex comments
  temp = regexpr("[^\\]%", tex)
  Loc = which(temp != -1)
  if (length(Loc)>0) tex[Loc] = substring(tex[Loc], 1, temp[Loc])
  # convert latex % to ordinary %
  tex = gsub("\\%", "%", tex, fixed=TRUE)
  
  ### general checks of appropriate structure ###
  # Reject multiple {shinyFoo}'s on the same line
  temp = sapply(gregexpr("[\\](begin|end)[[:blank:]]*[{]shiny[[:alpha:]]+[}]", tex), length)
  if (any(temp) > 1) {
    stop('@', temp[1], ifelse(length(temp)==1, '', paste(' (and', length(temp)-1, ' others)')),
         " cannot have multiple {shinyXXX}'s on the same line")
  }
  #
  # Reject text before {shinyFoo}'s on the same line
  temp = regexpr("[^[:blank:]]+[[:blank:]]*[[\\](begin|end)[[:blank:]]*[{]shiny[[:alpha:]]+[}]", tex)
  tempS = which(temp != -1)
  tempL = length(tempS)
  if (tempL > 0) {
    stop("@", tempS[1], ifelse(tempL ==1, '', paste(" (and", tempL-1, "others)")),
         " cannot have text before \\begin{shinyXXX} or \\end{shinyXXX} on the same line")
  }
  # Reject text after begin{shinyFoo}{bar} on the same line
  temp = str_match(tex, "[\\\\]begin[[:blank:]]*[{]shiny[[:alpha:]]+[}][[:blank:]]*([{].+[}])?[[:blank:]]*([^[:blank:]]+)")
  sel = !is.na(temp[,2])
  sel2 = substring(temp[sel,3], 1, 1) != "{"
  temp = sum(sel2)
  if (temp > 0) {
     stop("@", (1:length(tex))[sel][sel2][1], ifelse(temp==1, '', paste(" (and", temp-1, "others)")),
          " cannot have text after \\begin{shinyXXX}[{YYY}] on the same line")
  }
  
  # Reject text after end{shinyFoo} on the same line
  temp = regexpr("[[\\]end[[:blank:]]*[{]shiny[[:alpha:]]+[}][[:blank:]]*[^[:blank:]]", tex)
  tempS = which(temp != -1)
  tempL = length(tempS)
  if (tempL > 0) {
    stop("@", tempS[1], ifelse(tempL ==1, '', paste(" (and", tempL-1, "others)")),
         " cannot have text after \\end{shinyXXX} on the same line")
  }
  
  ### get "def"s and remove "def" lines  (note: allows more than one per line)  ##
  defs = gregexpr("[\\]def[[:blank:]]*[\\][[:alpha:]]+?[[:blank:]]*[{}].*?[}]", tex)
  defLines = which(sapply(defs, function(x)x[1]!=-1))
  defs = defs[defLines]
  if (length(defs)==0) {
    defs = vector("list", 0)
  } else {
    defs = lapply(1:length(defs), 
                  function(d) {
                    str = substring(tex[defLines[d]], 5) # drop "\def"
                    start=as.numeric(defs[[d]]) 
                    end=attr(defs[[d]],"match.length") - 4
                    pairs = lapply(1:length(start),
                                   function(e, s=str) {
                                     temp = substring(s, start[e], start[e]+end[e]-1)
                                     temp1 = regmatches(temp, regexpr("[[:alpha:]]+", temp))
                                     temp2 = regmatches(temp, regexpr("[{].*[}]", temp))
                                     temp2 = substring(temp2, 2, nchar(temp2)-1)
                                     return(cbind(name=temp1, value=temp2))
                                   })
                    return(do.call(rbind, pairs))
                  })
    defs = do.call(rbind, defs)
    defNames = defs[,"name"]
    defs = as.list(defs[,"value"])
    names(defs) = defNames
  }
  # 
  # Fill in defaults for missing defs
  if (is.null(defs$title)) defs$title = "untitled Shiny app"
  if (is.null(defs$name)) defs$name = ""
  if (is.null(defs$titleColor)) defs$titleColor = "red"
  if (is.null(defs$titleFontWeight)) defs$titleFontWeight = "bold"
  if (is.null(defs$titleFontPercent)) defs$titleFontPercent = "125"
  if (is.null(defs$titleFont)) defs$titleFont = "serif"
  #
  # Handle "shinythemes"
  if (!is.null(defs$theme)) {
    if (!requireNamespace("shinythemes")) stop("use of '\\def\\theme{}' requires installing package 'shinythemes'")
    themes = list.files(file.path(system.file(package="shinythemes"),"shinythemes","css"))
    themes = substring(themes, 1, nchar(themes)-8)
    if (!defs$theme %in% themes) {
      stop("theme '", defs$theme, "' is not valid; valid themes are ", paste(themes, collapse=", "))
    }
  }

  
  ### check for begin/end document and discard before and after ###
  begin = grep("[\\]begin[[:blank:]]*[{]document[}]", tex)
  if (length(begin)==0) stop("no \\begin{document}")
  if (length(begin)>1) stop("multiple \\begin{document}s")
  if (length(defLines)>0 && any(defLines>begin[1])) 
    stop("please move all \\def's above the begin{document} and try again")
  end = grep("[\\]end[[:blank:]]*[{]document[}]", tex)[1]
  if (length(end)==0) stop("no \\end{document}")
  if (length(end)>1) stop("multiple \\end{document}s")
  tex = tex[(begin+1):(end-1)]
  lineNumOffset = begin
  
  
  
  ### find shinyTab and shinySinglePage environments, and verify ###
  tabStart = grep("^[[:blank:]]*[\\]begin[[:blank:]]*[{]shinyTab[}]", tex)
  nTabs = length(tabStart) # number of tabs
  tabEnd = grep("^[[:blank:]]*[\\]end[[:blank:]]*[{]shinyTab[}]", tex)
  if (length(tabEnd) != nTabs) {
    stop(nTabs, " begin{shinyTab}s, but ", length(tabEnd), " end{shinyTab}s")
  }
  if (length(tabStart)>0) {
    mainStartEnd = cbind(tabStart, tabEnd)
    temp = diff(as.vector(t(mainStartEnd)))
    if (any(temp <=0 )) {
      temp = which(temp <=0 )[1]
      temp = as.vector(t(mainStartEnd))[c(temp,temp+1)] + lineNumOffset
      stop("@", temp[2], "/", temp[1], " intertwined begin/end{shinyTab}s")
    }
  #    
  } else { # Or handle a single page app
    # Assure appropriate use of \singlePageStartEnd
    singlePageStart = grep("[\\]begin[[:blank:]]*[{]shinySinglePage[}]", tex)
    singlePageEnd = grep("[\\]end[[:blank:]]*[{]shinySinglePage[}]", tex)
    if (length(singlePageStart)==0) stop("either \\begin{shinyTab} or \\begin{shinySinglePage} is required")
    if (length(singlePageEnd)==0) stop("missing \\end{shinySinglePage")
    if (length(singlePageStart)>1) {
      stop("@", singlePageStart[1]+lineNumOffset, "/", singlePageStart[2]+lineNumOffset,
           " only one \\begin{shinySinglePage} is allowed")
    }
    if (length(singlePageEnd)>1) {
      stop("@", singlePageEnd[1]+lineNumOffset, "/", singlePageEnd[2]+lineNumOffset,
           " only one \\end{shinySinglePage} is allowed")
    }
    if (singlePageEnd<=singlePageStart) {
      stop("@", singlePageEnd+lineNumOffset, "/", singlePageStart+lineNumOffset,
           " \\end{shinySinglePage} before \\begin{shinySinglePage}")
    }
    #
    # Assure appropriate (optional) use of 'shinySidebar' and 'shinyMain' environments
    sideLocB = grep("[\\]begin[[:blank:]]*[{]shinySidebar[}]", tex)
    sideLocE = grep("[\\]end[[:blank:]]*[{]shinySidebar[}]", tex)
    mainLocB = grep("[\\]begin[[:blank:]]*[{]shinyMain[}]", tex)
    mainLocE = grep("[\\]end[[:blank:]]*[{]shinyMain[}]", tex)
    if (length(sideLocB)>1) stop("@", paste(lineNumOffset+sideLocB, collapse=", "), " shinySinglePage allows only one \\begin{shinySidebar}")
    if (length(mainLocB)>1) stop("@", paste(lineNumOffset+mainLocB, collapse=", "), " shinySinglePage allows only one \\begin{shinyMain}")
    if (length(sideLocE)>1) stop("@", paste(lineNumOffset+sideLocE, collapse=", "), " shinySinglePage allows only one \\end{shinySidebar}")
    if (length(mainLocE)>1) stop("@", paste(lineNumOffset+mainLocE, collapse=", "), " shinySinglePage allows only one \\end{shinyMain}")
    if (length(sideLocB) != length(sideLocE)) stop("\\begin{shinySidebar} must be paired with \\end{shinySidebar}")
    if (length(mainLocB) != length(mainLocE)) stop("\\begin{shinyMain} must be paired with \\end{shinyMain}")
    if (length(sideLocB) != length(mainLocB)) 
      stop("'shinySidebar' and 'shinyMain' environments must be used together or not at all")
    #
    # Handle withSidebar vs. without
    if (length(sideLocB) > 0) {  # using withSidebar (sidebar + main)
      if (sideLocE <= sideLocB) stop("@", lineNumOffset+sideLocE, "/", lineNumOffset+sideLocB, " shinySidebar end before begin")
      if (mainLocE <= mainLocB) stop("@", lineNumOffset+mainLocE, "/", lineNumOffset+mainLocB, " shinyMain end before begin")
      if (sideLocB < mainLocB  && mainLocB < sideLocE) stop("@", lineNumOffset+mainLocB, "/", lineNumOffset+sideLocE, " intertwined shinySidebar and shinyMain")
      if (mainLocB < sideLocB  && sideLocB < mainLocE) stop("@", lineNumOffset+sideLocE, "/", lineNumOffset+mainLocB, " intertwined shinySidebar and shinyMain")
      mainStartEnd = rbind(c(sideLocB+1, sideLocE-1), c(mainLocB+1, mainLocE-1))
    } else { # just a single unformatted page
      mainStartEnd = rbind(c(singlePageStart+1, singlePageEnd-1))
    }
    nTabs = 0
  }
  #
  # verify only one type is used
  if (nTabs>0 && length(grep("[\\](begin|end)[[:blank:]](shinySinglePage|shinySidebar|shinyMain)",tex)>0)) {
    stop("cannot combine shinyTabs with shinySinglePage/shinySidebar/shinyMain")
  }

  
  ### Get tab titles ###
  if (nTabs==0) {
    tabTitles = NULL
  } else {
    tabTitles = rep("", nTabs)
    for (tab in 1:nTabs) {
      temp = regexpr("^[[:blank:]]*[\\]begin[[:blank:]]*[{]shinyTab[}][[:blank:]]*[{].*[}]", tex[tabStart[tab]])
      if (temp == -1) {
        txt = paste0("{", "Tab", tab, "}")
        warning("tab #", tab, " has no title")
      } else {
        txt = regmatches(tex[tabStart[tab]], temp)
      }
      temp = regexpr("[{][^}]*[}]$", txt)
      tabTitles[tab] = substring(txt, temp+1, temp+attr(temp,"match.length")-2)
    } # end for each tab
  } # end if tabs are being used
  
  ### warn of tex lines outside of shiny environments ###
  rCommentLines = grep("^[[:blank:]]*[#]", tex)
  blankLines = grep("^[[:blank:]]*$", tex)
  texLines = setdiff(1:length(tex), union(blankLines, rCommentLines))
  if (nTabs==0) 
    texLines = setdiff(texLines, c(singlePageStart, singlePageEnd, sideLocB, sideLocE, mainLocB, mainLocE))
  inTab = unlist(lapply(1:nrow(mainStartEnd), function(r) mainStartEnd[r,1]:mainStartEnd[r,2]))
  outside = setdiff(texLines, inTab)
  if (length(outside)>0)  {
    warning("lines outside of shiny environment blocks ignored ", "@", lineNumOffset+outside[1], 
            ifelse(length(outside)==1, "", paste(" plus", length(outside)-1, "more")))
  }

  return(list(cleanTex=tex, defs=defs, mainStartEnd=mainStartEnd, tabTitles=tabTitles, 
              tabbing=nTabs>0, lineNumOffset=lineNumOffset))
} # end parseTex()
